/* Demonstration program showing use of the Array package             */
/*                                                                    */
/* TESTBLAS: This program computes the matrix product C = A x B       */
/* using three different approaches:                                  */
/*    1. A single call to BLAS dgemm routine                          */
/*    2. One call to BLAS dgemv for each column of result matrix C    */
/*    3. One call to BLAS ddot for each element of result matrix C    */
/* Results in each case are checked for validity.                     */

import com.ibm.math.array.*;	// import the Array package

public class testblas {


    public static void print(doubleArray2D C) {

	try {
	    int m = C.size(0);
	    int p = C.size(1);

	    for (int i=0; i<m; i++) {
		if (i > 0) System.out.println();
		if (p > 0) System.out.print(C.get(i,0));
		for (int j=1; j<p; j++) {
		    System.out.print(" "+C.get(i,j));
		}
	    }
	    System.out.println();
	} catch (InvalidArrayAxisException e) {
	    System.out.println("Unexpected error in print, terminating ...");
	    System.exit(-1);
	}
    }


    public static boolean valid(doubleArray2D C, int n) {
	
	/*
	 * Check validity of results. Element C(i,j) should be 
	 * equal to i*n*(n-1)/2 (within an epsilon).
	 */
	double eps = 1e-09;

	try {
	    int m = C.size(0);
	    int p = C.size(1);

	    for (int i=0; i<m; i++) {
		for (int j=0; j<p; j++) {
		    double correct = ((double)i*(double)n*(double)(n-1)/2.0);
		    double error = Math.abs(C.get(i,j)-correct);
		    if (error > eps) {
			return false;
		    }
		}
	    }
	} catch (InvalidArrayAxisException e) {
	    System.out.println("Unexpected error in valid, terminating ...");
	    System.exit(-1);
	}

	return true;
    }
    
    public static void main(String[] args) 
	throws InvalidArrayShapeException, InvalidRangeException,
	       NonconformingArrayException, InvalidArrayAxisException {

	/*
	 * Default matrices are 64 x 64. Can be changed with
	 * command line arguments.
	 */
	int m = 128;
	int n = 128;
	int p = 128;
	boolean print = false;

	/*
	 * Parse command line arguments.
	 */
	try{
	    
	    for(int arg=0; arg<args.length; arg++) {
		if (args[arg].equals("-m")) 
		    m = Integer.parseInt(args[++arg]);
		else if (args[arg].equals("-n"))
		    n = Integer.parseInt(args[++arg]);
		else if (args[arg].equals("-p"))
		    p = Integer.parseInt(args[++arg]);
		else if (args[arg].equals("-print"))
		    print = true;
		else if (args[arg].equals("-h"))
		    throw new Exception();
		else if (args[arg].equals("-help"))
		    throw new Exception();
		else throw new Exception();
	    }
	} catch (Exception e) {
	    System.out.println("usage: testblas [-m <m>] [-n <n>] [-p <p>] [-print]");
	    return;
	}

	System.out.println("matrix multiply: "+m+" x "+n+" x "+p);

	double ops = 2.0*m*n*p;
	int reps = (int)(1.0e+08/ops);
	if (reps < 1) reps = 1;

	/*
	 * Initialize the matrices A(mxn) and B(nxp) so that
	 * row i of each has all elements equal to i.
	 */
	doubleArray2D A = new doubleArray2D(m,n);
	doubleArray2D B = new doubleArray2D(n,p);
	for (int i=0; i<m; i++) {
	    A.section(i,new Range(0,n-1)).assign(i);
	}
	for (int i=0; i<n; i++) {
	    B.section(i,new Range(0,p-1)).assign(i);
	}
	doubleArray2D C = new doubleArray2D(m,p);

	/*
	 * Use variable etime to measure elapsed time, mflops to
	 * compute Mflops.
	 */
	double etime = 0;
	double mflops = 0;

	/*
	 * Perform the matrix multiplication of A and B, yielding C.
	 * Use Blas.dgemm.
	 */
	etime = System.currentTimeMillis();
	for (int r=0; r<reps; r++) {
	    Blas.dgemm(Blas.NoTranspose,Blas.NoTranspose,1.0,A,B,0.0,C);
	}
	etime = System.currentTimeMillis() - etime;

	/*
	 * Print performance.
	 */
	mflops = 1.0e-3*(2.0*m*n*p*reps)/etime;
	System.out.print("mflops from dgemm: "+mflops);

	/*
	 * Check validity of result.
	 */
	if (valid(C,n)) {
	    System.out.println(" Result is correct!");
	} else {
	    System.out.println(" Result is incorrect!");
	}

	/*
	 * If print flag is set, print the result.
	 */
	if (print) {
	    System.out.println("C matrix from dgemm:");
	    print(C);
	}

	/*
	 * Perform the matrix multiplication of A and B, yielding C.
	 * Use Blas.dgemv.
	 */
	etime = System.currentTimeMillis();
	for (int r=0; r<reps; r++) {
	    for (int j=0; j<p; j++) {
		Blas.dgemv(Blas.NoTranspose,1.0,A,
			   B.section(new Range(0,n-1),j),
			   0.0,C.section(new Range(0,m-1),j));
	    }
	}
	etime = System.currentTimeMillis() - etime;

	/*
	 * Print performance.
	 */
	mflops = 1.0e-3*(2.0*m*n*p*reps)/etime;
	System.out.print("mflops from dgemv: "+mflops);

	/*
	 * Check validity of result.
	 */
	if (valid(C,n)) {
	    System.out.println(" Result is correct!");
	} else {
	    System.out.println(" Result is incorrect!");
	}

	/*
	 * If print flag is set, print the result.
	 */
	if (print) {
	    System.out.println("C matrix from dgemv:");
	    print(C);
	}

	/*
	 * Perform the matrix multiplication of A and B, yielding C.
	 * Use Blas.ddot.
	 */
	etime = System.currentTimeMillis();
	for (int r=0; r<reps; r++) {
	    for (int i=0; i<m; i++) {
		for (int j=0; j<p; j++) {
		    C.set(i,j,Blas.ddot(A.section(i,new Range(0,n-1)),
					B.section(new Range(0,n-1),j)));
		}
	    }
	}
	etime = System.currentTimeMillis() - etime;

	/*
	 * Print performance.
	 */
	mflops = 1.0e-3*(2.0*m*n*p*reps)/etime;
	System.out.print("mflops from ddot: "+mflops);

	/*
	 * Check validity of result.
	 */
	if (valid(C,n)) {
	    System.out.println(" Result is correct!");
	} else {
	    System.out.println(" Result is incorrect!");
	}

	/*
	 * If print flag is set, print the result.
	 */
	if (print) {
	    System.out.println("C matrix from ddot:");
	    print(C);
	}
    }
}
